/*
 * PeTrack - Software for tracking pedestrians movement in videos
 * Copyright (C) 2010-2021 Forschungszentrum Jülich GmbH,
 * Maik Boltes, Juliane Adrian, Ricardo Martin Brualla, Arne Graf, Paul Häger, Daniel Hillebrand,
 * Deniz Kilic, Paul Lieberenz, Daniel Salden, Tobias Schrödter, Ann Katrin Seemann
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "trackerReal.h"

#include "helper.h"
#include "personStorage.h"
#include "petrack.h"
#include "recognition.h"

TrackPointReal::TrackPointReal(const Vec3F &p, int frameNum) : Vec3F(p), mFrameNum(frameNum)
{
    mAngleOfView = -1;
    mMarkerID    = -1;
}

TrackPointReal::TrackPointReal(const Vec3F &p, int frameNum, const Vec2F &d) :
    Vec3F(p), mFrameNum(frameNum), mViewDir(d)
{
    mAngleOfView = -1;
    mMarkerID    = -1;
}

//--------------------------------------------------------------------------

// the list index is the frame number plus mFirstFrame 0..mLastFrame-mFirstFrame
// no frame is left blank
TrackPersonReal::TrackPersonReal() : mHeight(-1.), mFirstFrame(-1), mLastFrame(0) {}
TrackPersonReal::TrackPersonReal(int frame, const TrackPointReal &p) :
    mHeight(-1.), mFirstFrame(frame), mLastFrame(frame)
{
    append(p);
}

bool TrackPersonReal::trackPointExist(int frame) const
{
    if(frame >= mFirstFrame && frame <= mLastFrame)
    {
        return true;
    }
    else
    {
        return false;
    }
}
const TrackPointReal &
TrackPersonReal::trackPointAt(int frame) const // & macht bei else probleme, sonst mit [] zugreifbar
{
    return at(frame - mFirstFrame);
}

// gibt -1 zurueck, wenn frame oder naechster frame nicht existiert
// entfernung ist absolut
double TrackPersonReal::distanceToNextFrame(int frame) const
{
    if(frame >= mFirstFrame && frame + 1 <= mLastFrame)
    {
        return at(frame - mFirstFrame).distanceToPoint(at(frame - mFirstFrame + 1));
    }
    else
    {
        return -1;
    }
}
void TrackPersonReal::init(int firstFrame, double height, int markerID)
{
    clear();
    mFirstFrame = firstFrame;
    mLastFrame  = firstFrame - 1;
    mHeight     = height;
    mMarkerID   = markerID;
}

void TrackPersonReal::addEnd(const QPointF &pos, int frame)
{
    Vec3F point(pos.x(), pos.y(), -1.);
    addEnd(point, frame);
}
void TrackPersonReal::addEnd(const Vec3F &pos, int frame)
{
    ++mLastFrame;
    append(TrackPointReal(pos, frame));
}

void TrackPersonReal::addEnd(const QPointF &pos, int frame, const QPointF &dir)
{
    Vec3F point(pos.x(), pos.y(), -1.);
    Vec2F viewDirection(dir.x(), dir.y());
    ++mLastFrame;
    viewDirection.normalize();
    append(TrackPointReal(point, frame, viewDirection));
}

//-------------------------------------------------------------------------------------------------------------------------------------

TrackerReal::TrackerReal(QWidget *wParent, PersonStorage &storage) :
    mXMin(100000.), mXMax(-100000.), mYMin(100000.), mYMax(-100000.), mPersonStorage(storage)
{
    mMainWindow = (class Petrack *) wParent;
}

// default: int imageBorderSize = 0, bool missingFramesInserted = true, bool useTrackpoints = false
int TrackerReal::calculate(
    Tracker *  tracker,
    ImageItem *imageItem,
    ColorPlot *colorPlot,
    int        imageBorderSize,
    bool       missingFramesInserted,
    bool       useTrackpoints,
    bool       alternateHeight,
    double     altitude,
    bool       useCalibrationCenter,
    bool       exportElimTp,
    bool       exportElimTrj,
    bool       exportSmooth,
    bool       exportViewingDirection,
    bool       exportAngleOfView,
    bool       exportMarkerID,
    bool       exportAutoCorrect)
{
    if(tracker || imageItem || colorPlot)
    {
        if(size() > 0)
        {
            clear();
        }

        int        j, f;
        QList<int> missingList;    // frame nr wo ausgelassen; passend dazu:
        QList<int> missingListAnz; // anzahl ausgelassener frames
        if(missingFramesInserted)
        {
            // finden von nicht aufgenommenen frames
            int                               largestLastFrame = mPersonStorage.largestLastFrame();
            QMap<int, double>                 distanceMap;     // map der distanzen zu frame
            QMap<int, double>                 lastDistanceMap; // map der distanzen zu frame
            QMap<int, double>::const_iterator iter;
            int                               anz, skipAnz;
            double                            dist, sum, lastSum = 0.;
            // (median neben durchschnittswert als bewertungszahl hinzugenommen,
            // da sonst bei wenigen personen und langsamen bewegungen nur ein kleiner ausreisser
            // einen frame hinzufuegt (insb linienexperiment mit vielen personen))
            for(f = mPersonStorage.smallestFirstFrame(); f <= largestLastFrame; ++f)
            {
                distanceMap.clear();
                skipAnz = anz = 0;
                sum           = 0.;
                // erzeugen distanz-map im aktuellen frame
                for(size_t i = 0; i < mPersonStorage.nbPersons(); ++i)
                {
                    if((dist = mPersonStorage.at(i).distanceToNextFrame(f)) > -1)
                    {
                        distanceMap[static_cast<int>(i)] = dist;
                    }
                }
                // distanzliste mit vorherigem frame vergleichen
                iter = distanceMap.constBegin();
                while(iter != distanceMap.constEnd())
                {
                    if(lastDistanceMap.contains(iter.key()) &&
                       lastDistanceMap[iter.key()] > 1.) // damit bei ganz kleinen bewegungen nicht angeschlagen wird
                    {
                        ++anz;
                        if(distanceMap[iter.key()] > 1.5 * lastDistanceMap[iter.key()])
                        {
                            ++skipAnz;
                        }
                    }
                    sum += distanceMap[iter.key()];
                    ++iter;
                }
                sum /= distanceMap.size();
                // bei sprung in mehreren trackingpfaden als sprung markieren
                if((anz > 1) && (skipAnz / (double) anz > .5) && myRound(sum / lastSum - 1.) > 0) // 50% (war (anz > 2))
                {
                    missingList.append(f);
                    missingListAnz.append(myRound(sum / lastSum - 1.));
                    debout << "Warning: potentially missing " << myRound(sum / lastSum - 1.) << " frame(s) between "
                           << f << " and " << f + 1 << " will be inserted." << std::endl;
                }
                lastDistanceMap = distanceMap;
                lastSum         = sum;
            }
        }

        // fps ist nicht aussagekraeftig, da sie mgl von ausgelassenen herruehren - besser immer 25,01 fps annehmen

        double          height; // groesse in cm
        int             firstFrame, addFrames, anz;
        Vec2F           br(imageBorderSize, imageBorderSize);
        QPointF         pos, pos2;
        QList<int>      tmpMissingList;    // frame nr
        QList<int>      tmpMissingListAnz; // anzahl frames
        TrackPersonReal trackPersonReal;
        QPointF         center = imageItem->getPosReal(
            QPointF(imageItem->boundingRect().width() / 2., imageItem->boundingRect().height() / 2.), 0.);
        Vec3F   sp;
        int     tsize;
        int     extrapolated;
        QPointF colPos;
        float   angle;

        const auto &persons = mPersonStorage.getPersons();
        for(size_t i = 0; i < persons.size(); ++i) // ueber trajektorien
        {
            const auto &person = persons[i];
            addFrames          = 0;
            firstFrame         = person.firstFrame();

            if(person.height() < MIN_HEIGHT + 1)
            {
                height = colorPlot->map(person.color());
            }
            else
            {
                height = person.height();
            }

            if(missingList.size() > 0)
            {
                tmpMissingList    = missingList;
                tmpMissingListAnz = missingListAnz;
                // vorspulen
                while((tmpMissingList.size() > 0) && (tmpMissingList.first() < firstFrame))
                {
                    tmpMissingList.removeFirst();               // frame
                    addFrames += tmpMissingListAnz.takeFirst(); // anzahl
                }
            }

            int markerID = person.getMarkerID();

            trackPersonReal.init(firstFrame + addFrames, height, markerID);
            tsize = person.size();
            for(j = 0; (j < tsize); ++j) // ueber trackpoints
            {
                // ausreisser ausfindig machen (dies geschieht, bevor -1 elemente herausgenommen werden, um die
                // glaettung beim eliminieren der -1 elemente hier nicht einfluss nehmen zu lassen):
                if(exportSmooth && (useTrackpoints || alternateHeight) &&
                   (tsize > 1)) // wenn direkt pointgrey hoehe oder eigene hoehenberechnung aber variierend ueber trj
                                // genommen werden soll
                {
                    // changes Trajectories!
                    mPersonStorage.smoothHeight(i, j);
                }
                if(useTrackpoints)
                {
                    // border unberuecksichtigt
                    if(useCalibrationCenter)
                    {
                        trackPersonReal.addEnd(
                            Vec3F(
                                person[j].sp().x() + center.x(),
                                center.y() - person[j].sp().y(),
                                altitude - person[j].sp().z()),
                            firstFrame + j);
                    }
                    else
                    {
                        trackPersonReal.addEnd(person[j].sp(), firstFrame + j);
                    }
                }
                else
                {
                    if(alternateHeight) // personenhoehe variiert ueber trajektorie (unebene versuche); berechnung durch
                                        // mich und nicht pointgrey nutzen, Kamera altitude nutzen
                    {
                        extrapolated = 0; // 0 == false
                        double bestZ =
                            person.getNearestZ(j, &extrapolated); // gibt z wert zurueck bzw wenn -1, dann den
                                                                  // neben diesem frame ersten z-wert ungleich -1
                        if(bestZ < 0)                             // == -1 // es liegt gar keine berechnete hoehe vor
                        {
                            if(exportElimTrj)
                            {
                                debout << "Warning: no calculated height for trackpoint " << j << " (frame "
                                       << person.firstFrame() + j << ") of person " << i + 1
                                       << ", person is not exported!" << std::endl;
                                break; // TrackPerson ist angelegt, erhaelt aber keine Points und wird deshalb am ende
                                       // nicht eingefuegt
                            }
                            else
                            {
                                bestZ = height; // wenn gar kein trackpoint ungleich -1 gefunden wird, dann wird zu
                                                // allerletzt Hoehe genommen
                                debout << "Warning: no calculated height for trackpoint " << j << " (frame "
                                       << person.firstFrame() + j << ") of person " << i + 1
                                       << ", default height is used!" << std::endl;
                            }
                        }
                        else
                        {
                            bestZ = altitude - bestZ;
                        }
                        // ab hier kann bestZ auch negativ sein, obwohl hoehe inhalt hatte

                        if(extrapolated && exportElimTp)
                        {
                            if(extrapolated == 1) // zu beginn verworfen
                            {
                                trackPersonReal.setFirstFrame(trackPersonReal.firstFrame() + 1);
                            }
                            debout << "Warning: no calculated height for trackpoint " << j << " (frame "
                                   << person.firstFrame() + j << ") of person " << i + 1
                                   << ", extrapolated height not used, trackpoint not inserted!" << std::endl;
                        }
                        else
                        {
                            if(extrapolated)
                            {
                                debout << "Warning: no calculated height for trackpoint " << j << " (frame "
                                       << person.firstFrame() + j << ") of person " << i + 1
                                       << ", extrapolated height is used!" << std::endl;
                            }

                            Vec2F moveDir(0, 0);
                            if(exportAutoCorrect)
                            {
                                moveDir += reco::autoCorrectColorMarker(person[j], mMainWindow->getControlWidget());
                            }

                            pos = imageItem->getPosReal((person[j] + moveDir + br).toQPointF(), bestZ);

                            if((exportViewingDirection) &&
                               (person[j].color().isValid())) // wenn blickrichtung mit ausgegeben werden soll
                            {
                                colPos =
                                    imageItem->getPosReal((person[j].colPoint() + moveDir + br).toQPointF(), bestZ);
                                trackPersonReal.addEnd(pos, firstFrame + j, colPos - pos);
                            }
                            else
                            {
                                trackPersonReal.addEnd(Vec3F(pos.x(), pos.y(), bestZ), firstFrame + j);
                            }
                        }
                    }
                    else
                    {
                        Vec2F moveDir(0, 0);
                        if(exportAutoCorrect)
                        {
                            moveDir += reco::autoCorrectColorMarker(person[j], mMainWindow->getControlWidget());
                        }

                        pos = imageItem->getPosReal((person[j] + moveDir + br).toQPointF(), height);
                        // die frame nummer der animation wird TrackPoint der PersonReal mitgegeben,
                        // da Index groesser sein kann, da vorher frames hinzugefuegt wurden duch
                        // trackPersonReal.init(firstFrame+addFrames, height) oder aber innerhalb des trackink path mit
                        // for schleife ueber f
                        if((exportViewingDirection) &&
                           (person[j].color().isValid())) // wenn blickrichtung mit ausgegeben werden soll
                        {
                            colPos = imageItem->getPosReal((person[j].colPoint() + moveDir + br).toQPointF(), height);
                            trackPersonReal.addEnd(pos, firstFrame + j, colPos - pos);
                        }
                        else
                        {
                            trackPersonReal.addEnd(pos, firstFrame + j);
                        }
                        if(exportAngleOfView)
                        {
                            angle = (90. -
                                     imageItem->getAngleToGround((person[j] + br).x(), (person[j] + br).y(), height)) *
                                    PI / 180.;
                            trackPersonReal.last().setAngleOfView(angle);
                        }
                        if(exportMarkerID)
                        {
                            trackPersonReal.last().setMarkerID(person[j].getMarkerID());
                        }
                    }
                }

                if(tmpMissingList.size() > 0)
                {
                    if((tmpMissingList.first() == firstFrame + j) && (person.trackPointExist(firstFrame + j + 1)))
                    {
                        tmpMissingList.removeFirst();        // frame
                        anz = tmpMissingListAnz.takeFirst(); // anzahl
                        if(useTrackpoints)
                        {
                            // border unberuecksichtigt
                            for(f = 1; f <= anz; ++f)
                            {
                                sp = person[j].sp() + f * (person[j + 1].sp() - person[j].sp()) / (anz + 1);
                                if(useCalibrationCenter)
                                {
                                    trackPersonReal.addEnd(
                                        Vec3F(sp.x() + center.x(), center.y() - sp.y(), altitude - sp.z()),
                                        firstFrame + j);
                                }
                                else
                                {
                                    trackPersonReal.addEnd(sp, -1); // -1 zeigt an, dass nur interpoliert
                                }
                            }
                        }
                        else
                        {
                            if(alternateHeight) // personenhoehe variiert ueber trajektorie (unebene versuche);
                                                // berechnung durch mich und nicht pointgrey nutzen
                            {
                                debout << "Warning: No interpolation is done, because alternate height is enabled - "
                                          "has to be implemented!"
                                       << std::endl;
                            }
                            else
                            {
                                Vec2F moveDir(0, 0);
                                if(exportAutoCorrect)
                                {
                                    moveDir += reco::autoCorrectColorMarker(person[j], mMainWindow->getControlWidget());
                                }

                                pos2 =
                                    (imageItem->getPosReal((person[j + 1] + moveDir + br).toQPointF(), height) - pos) /
                                    (anz + 1);
                                for(f = 1; f <= anz; ++f)
                                {
                                    trackPersonReal.addEnd(pos + f * pos2, -1); // -1 zeigt an, dass nur interpoliert
                                }
                            }
                        }
                    }
                    else if(tmpMissingList.first() < firstFrame) // while, wenn nicht kontinuierlich waere
                    {
                        tmpMissingList.removeFirst();    // frame
                        tmpMissingListAnz.removeFirst(); // anzahl
                    }
                }
            }

            tsize                  = trackPersonReal.size();
            double maxHeightDiff   = 30.; // 30cm
            int    numBorderPoints = 50;
            int    k;
            int    delNumFront = 0;
            // die ersten und letzten numBorderPoints trackpoints untersuchen, ob die hoehe einen sprung groesser
            // maxHeightDiff aufweist
            for(j = 1; j < tsize; ++j) // ueber trackpoints of personreal; ab 1, da vergleich zu vorherigem
            {
                if((j < numBorderPoints) &&
                   ((trackPersonReal.at(j).z() - trackPersonReal.at(j - 1).z()) > maxHeightDiff))
                {
                    delNumFront = j;
                }
                else if(
                    ((tsize - numBorderPoints) < j) &&
                    ((trackPersonReal.at(j).z() - trackPersonReal.at(j - 1).z()) > maxHeightDiff))
                {
                    for(k = j; k < tsize; k++)
                    {
                        trackPersonReal.setLastFrame(trackPersonReal.lastFrame() - 1);
                        trackPersonReal.removeLast();
                    }
                    debout << "Warning: delete last " << tsize - j << " frames of person " << i + 1
                           << " because of height jump!" << std::endl;
                    break;
                }
            }
            tsize = trackPersonReal.size();
            if(delNumFront > 0)
            {
                debout << "Warning: delete first " << delNumFront << " frames of person " << i + 1
                       << " because of height jump!" << std::endl;
                for(k = 0; (k < delNumFront) && (k < tsize); k++)
                {
                    trackPersonReal.setFirstFrame(trackPersonReal.firstFrame() + 1);
                    trackPersonReal.removeFirst();
                }
            }
            if(trackPersonReal.size() < 20)
            {
                debout << "Warning: Person " << i + 1 << " has only " << trackPersonReal.size() << " trackpoints!"
                       << std::endl;
            }
            if(trackPersonReal.size() > 0)
            {
                append(trackPersonReal);
            }
            else // ggf weil keine calculated height vorlag (siehe exportElimTrj)
            {
                debout << "Warning: Person " << i + 1 << " is not inserted, because of no trackpoints!" << std::endl;
            }
        }
        return size();
    }
    else
    {
        debout << "Warning: no real tracking data calculated, because of missing tracking data, image reference or "
                  "color map!"
               << std::endl;
        return -1;
    }
}

void TrackerReal::calcMinMax()
{
    Vec3F pos;
    for(int i = 0; i < size(); ++i)
    {
        for(int j = 0; j < at(i).size(); ++j)
        {
            pos = at(i).at(j);
            if(mXMin > pos.x())
            {
                mXMin = pos.x();
            }
            else if(mXMax < pos.x())
            {
                mXMax = pos.x();
            }
            if(mYMin > pos.y())
            {
                mYMin = pos.y();
            }
            else if(mYMax < pos.y())
            {
                mYMax = pos.y();
            }
        }
    }
}

int TrackerReal::largestFirstFrame()
{
    int max = -1, i;
    for(i = 0; i < size(); ++i)
    {
        if(at(i).firstFrame() > max)
        {
            max = at(i).firstFrame();
        }
    }
    return max;
}
int TrackerReal::largestLastFrame()
{
    int max = -1, i;
    for(i = 0; i < size(); ++i)
    {
        if(at(i).lastFrame() > max)
        {
            max = at(i).lastFrame();
        }
    }
    return max;
}
int TrackerReal::smallestFirstFrame()
{
    int i, min = ((size() > 0) ? at(0).firstFrame() : -1);
    for(i = 1; i < size(); ++i)
    {
        if(at(i).firstFrame() < min)
        {
            min = at(i).firstFrame();
        }
    }
    return min;
}
int TrackerReal::smallestLastFrame()
{
    int i, min = ((size() > 0) ? at(0).lastFrame() : -1);
    for(i = 1; i < size(); ++i)
    {
        if(at(i).lastFrame() < min)
        {
            min = at(i).lastFrame();
        }
    }
    return min;
}

void TrackerReal::exportTxt(
    QTextStream &out,
    bool         alternateHeight,
    bool         useTrackpoints,
    bool         exportViewingDirection,
    bool         exportAngleOfView,
    bool         exportUseM,
    bool         exportMarkerID)
{
    float scale;

    out << "# z: can be 3d position or height of person (alternating or not)" << Qt::endl;
    if(exportViewingDirection)
    {
        out << "# viewDirX viewDirY: vector of direction of head of person" << Qt::endl;
    }
    if(exportAngleOfView)
    {
        out << "# viewAngle: angle of view of camera to person from perpendicular [0..Pi/2]" << Qt::endl;
    }
    if(exportUseM)
    {
        out << "# id frame x/m y/m z/m";
    }
    else
    {
        out << "# id frame x/cm y/cm z/cm";
    }
    if(exportViewingDirection)
    {
        out << " viewDirX viewDirY";
    }
    if(exportAngleOfView)
    {
        out << " viewAngle" << Qt::endl;
    }
    if(exportMarkerID)
    {
        out << " markerID" << Qt::endl;
    }
    else
    {
        out << Qt::endl;
    }

    if(exportUseM)
    {
        scale = .01f;
    }
    else
    {
        scale = 1.f;
    }

    QProgressDialog progress("Export TXT-File", nullptr, 0, size(), mMainWindow->window());
    progress.setWindowTitle("Export .txt-File");
    progress.setWindowModality(Qt::WindowModal);
    progress.setVisible(true);
    progress.setValue(0);
    progress.setLabelText(QString("Export tracking data ..."));
    qApp->processEvents();

    for(int i = 0; i < size(); ++i)
    {
        qApp->processEvents();
        progress.setLabelText(QString("Export person %1 of %2 ...").arg(i + 1).arg(size()));
        progress.setValue(i + 1);

        for(int j = 0; j < at(i).size(); ++j)
        {
            out << i + 1 << " " << at(i).firstFrame() + j << " " << at(i).at(j).x() * scale << " "
                << at(i).at(j).y() * scale << " ";

            if(alternateHeight || useTrackpoints)
            {
                out << at(i).at(j).z() * scale;
            }
            else
            {
                out << at(i).height() * scale;
            }

            if(exportViewingDirection) // && (at(i).at(j).viewDir() != Vec2F(0,0)) zeigt an, dass keine richtung
                                       // berechnet werden konnte
            {
                out << " " << at(i).at(j).viewDir();
            }

            if(exportAngleOfView)
            {
                out << " " << at(i).at(j).angleOfView() << Qt::endl;
            }

            if(exportMarkerID)
            {
                out << " " << at(i).getMarkerID() << Qt::endl;
            }
            else
            {
                out << Qt::endl;
            }
        }
    }
}

// old - not all export options supported!!!!
void TrackerReal::exportDat(QTextStream &out, bool alternateHeight, bool useTrackpoints) // fuer gnuplot
{
    QProgressDialog progress("Export DAT-File", nullptr, 0, size(), mMainWindow->window());
    progress.setWindowTitle("Export .dat-File");
    progress.setWindowModality(Qt::WindowModal);
    progress.setVisible(true);
    progress.setValue(0);
    progress.setLabelText(QString("Export tracking data ..."));
    debout << "size:" << size() << std::endl;
    if(useTrackpoints)
    {
        for(int i = 0; i < size(); ++i)
        {
            progress.setLabelText(QString("Export person %1 of %2 ...").arg(i + 1).arg(size()));
            qApp->processEvents();
            progress.setValue(i + 1);

            for(int j = 0; j < at(i).size(); ++j)
            {
                // Umrechnung in coordSystem fehlt, auch camera altitude unberuecksichtigt!!!
                out << at(i).firstFrame() + j << " " << at(i).at(j) << " " << at(i).at(j).z()
                    << Qt::endl; // z Koordinate ist Kopf
            }
            out << Qt::endl;
        }
    }
    else
    {
        for(int i = 0; i < size(); ++i)
        {
            progress.setLabelText(QString("Export person %1 of %2 ...").arg(i + 1).arg(size()));
            qApp->processEvents();
            progress.setValue(i + 1);
            debout << "person: " << i << std::endl;

            for(int j = 0; j < at(i).size(); ++j)
            {
                if(alternateHeight)
                {
                    out << at(i).firstFrame() + j << " " << at(i).at(j) << " " << at(i).at(j).z()
                        << Qt::endl; // z Koordinate ist Kopf
                }
                else
                {
                    out << at(i).firstFrame() + j << " " << at(i).at(j) << " " << at(i).height()
                        << Qt::endl; // z Koordinate ist Kopf
                }
            }
            out << Qt::endl;
        }
    }
}

void TrackerReal::exportXml(QTextStream &outXml, bool alternateHeight, bool useTrackpoints)
{
    int    i, j, largestLastFr = largestLastFrame();
    double z;
    int    defaultPersonHeight = 176;

    QProgressDialog progress("Export XML-File", nullptr, 0, largestLastFr, mMainWindow->window());
    progress.setWindowTitle("Export .xml-File");
    progress.setWindowModality(Qt::WindowModal);
    progress.setVisible(true);
    progress.setValue(0);
    progress.setLabelText(QString("Export tracking data ..."));
    qApp->processEvents();

    outXml << "    <shape>" << Qt::endl;
    for(j = 0; j < size(); ++j)
    {
        if(alternateHeight) // bei variierender groesse wird einfach durchschnittsgroesse genommen, da an treppen gar
                            // keine vernuempftige Groesse vorliegt
        {
            outXml << "        <agentInfo ID=\"" << j + 1 << "\" color=\"100\" height=\"" << defaultPersonHeight
                   << "\"/>" << Qt::endl;
        }
        else
        {
            outXml << "        <agentInfo ID=\"" << j + 1 << "\" color=\"100\" height=\"" << at(j).height() << "\"/>"
                   << Qt::endl;
        }
    }
    outXml << "    </shape>" << Qt::endl << Qt::endl;

    // i = frame; j = person
    for(i = smallestFirstFrame(); i <= largestLastFr; ++i)
    {
        progress.setLabelText(QString("Export data ..."));
        qApp->processEvents();
        progress.setValue(i);

        outXml << "    <frame ID=\"" << i << "\">" << Qt::endl;
        for(j = 0; j < size(); ++j)
        {
            if(at(j).trackPointExist(i))
            {
                // z-wert ist hier ausnahmsweise nicht der kopf, sondern der boden, die prsonengroesse wird dem obigem
                // person-datenentnommen
                // personID, Frame ID(?) , X , Y , Z
                if(useTrackpoints)
                {
                    z = at(j).trackPointAt(i).z() + defaultPersonHeight;
                }
                else // war, wenn ebene versuche sinnvoll:z = at(j).trackPointAt(i).z()+at(j).height();
                {
                    if(alternateHeight)
                    {
                        z = at(j).trackPointAt(i).z() - defaultPersonHeight;
                    }
                    else // war, wenn ebene versuche sinnvoll: z = at(j).trackPointAt(i).z()-at(j).height();
                    {
                        z = 0; // at(j).height();
                    }
                }
                outXml << "        <agent ID=\"" << j + 1 << "\" xPos=\"" << at(j).trackPointAt(i).x() << "\" yPos=\""
                       << at(j).trackPointAt(i).y() << "\" zPos=\"" << z
                       << "\" xVel=\"0\" yVel=\"0\" zVel=\"0\" radiusA=\"18\" radiusB=\"15\" "
                          "ellipseOrientation=\"130\" ellipseColor=\"0\"/>"
                       << Qt::endl; // z Koordinate ist Boden
            }
        }
        outXml << "    </frame>" << Qt::endl;
    }
}
